# This file is part of Arduino_Core_Rudiron_MDR32F9Qx.

# Arduino_Core_Rudiron_MDR32F9Qx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Arduino_Core_Rudiron_MDR32F9Qx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Arduino_Core_Rudiron_MDR32F9Qx. If not, see <https://www.gnu.org/licenses/>.

# Author: Daniil Ignatev. 23.09.2023
# Rudiron Distribution class
# Modified in 2024 by Alexandr Safonov for AQUARIUS_Arduino_Core_Rudiron_MDR32F9Qx

import os
import pathlib
import platform
import urllib.request
import ssl
from urllib.error import URLError, HTTPError
import tempfile
import shutil
import json
import hashlib
import zipfile
import tarfile
import subprocess


class Rudiron_Distribution:
    def __init__(self) -> None:
        self.__temp_folder_path = None
        self.__temp_maintainer_path = None
        self.__temp_hardware_path = None
        self.__temp_tools_path = None
        self.__source_url = "https://gitflic.ru/project/akvarius-rudiron/rudiron-distr/blob/raw?file=setup%2F"
        self.__index_filename = "package_Rudiron_index.json"
        self.__install_error = ""
        self.__vscode_extentions = [
            "ms-vscode.cpptools-extension-pack",
            "franmastromarino.vs-code-settings-os",
            "marus25.cortex-debug",
	    self.__get_vscode_extention_path()
        ]
        self.percentage_deci_last = 0

    def __get_vscode_extention_path(self):
        """Returns the path to Rudiron vscode_extention directory"""
        os_type = platform.system()

        if os_type == "Windows":
            return os.path.join(os.getenv("LOCALAPPDATA"), "Arduino15", "packages", "Rudiron", "tools", "rudiron-tasks", "default", "rudiron-tasks-1.0.0.vsix")
        elif os_type == "Darwin":
            return os.path.join(
                os.path.expanduser("~"), "Library", "Arduino15", "packages", "Rudiron", "tools", "rudiron-tasks", "default", "rudiron-tasks-1.0.0.vsix")
        elif os_type == "Linux":
            return os.path.join(os.path.expanduser("~"), ".arduino15", "packages", "Rudiron", "tools", "rudiron-tasks", "default", "rudiron-tasks-1.0.0.vsix")
        else:
            raise ValueError("Unsupported operating system")

    class Archive_Metadata:
        def __init__(
            self, archiveFileName, url, checksum, size, destination_directory
        ) -> None:
            self.archiveFileName = archiveFileName
            self.url = url
            self.checksum = checksum
            self.size = size
            self.destination_directory = destination_directory

    def __download_progress(self, count, block_size, total_size):
        downloaded = count * block_size
        downloaded = min(downloaded, total_size)

        percentage = (downloaded / total_size) * 100.0
        percentage_deci = int(percentage / 10)

        if percentage_deci != self.percentage_deci_last:
            self.percentage_deci_last = percentage_deci
            print("Downloaded {} of {} bytes ({}%)".format(downloaded, total_size, percentage_deci * 10))

    def __download_file(self, url, destination):
        error = None

        print("Downloading {}".format(url))

        try:
            index_json_path, m = urllib.request.urlretrieve(url, destination, reporthook=self.__download_progress)
        except HTTPError as e:
            error = "HTTP Error: {}. Reason: {}".format(e.code, e.reason)
        except URLError as e:
            if isinstance(e.reason, ssl.SSLError):
                error = "SSL Error: {}"
            else:
                error = "URL Error: {}".format(e.reason)
        except Exception as e:
            error = "Unexpected error: {}".format(e)

        if error != None:
            self.set_install_error(
                "Error downloading {} into {}. Details: {}".format(url, destination, error)
            )
    
    def __get_arduino_packages_path(self):
        """Returns the path to Arduino packages directory"""
        os_type = platform.system()

        if os_type == "Windows":
            return os.path.join(os.getenv("LOCALAPPDATA"), "Arduino15", "packages")
        elif os_type == "Darwin":
            return os.path.join(
                os.path.expanduser("~"), "Library", "Arduino15", "packages"
            )
        elif os_type == "Linux":
            return os.path.join(os.path.expanduser("~"), ".arduino15", "packages")
        else:
            raise ValueError("Unsupported operating system")

    def __get_arduino_maintainer_path(self):
        packages_path = self.__get_arduino_packages_path()
        maintainer_path = os.path.join(packages_path, "Rudiron")

        return maintainer_path

    def is_installed(self):
        """Return True if Rudiron distribution is already installed, otherwise False"""
        return os.path.exists(self.__get_arduino_maintainer_path())

    def get_install_error(self):
        return self.__install_error

    def has_error(self):
        return self.get_install_error() != ""

    def set_install_error(self, new_value):
        self.__install_error = new_value

    def __get_sha256_checksum(self, file_path):
        """Compute the SHA256 checksum of a file."""
        sha256_hash = hashlib.sha256()

        with open(file_path, "rb") as f:
            # Read and update hash in 4K blocks
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)

        return sha256_hash.hexdigest()

    def __check_archive(self, file_path, expected_size, expected_checksum):
        """Check the size and SHA256 checksum of an archive."""
        actual_size = os.path.getsize(file_path)
        actual_checksum = self.__get_sha256_checksum(file_path)

        is_size_valid = str(actual_size) == expected_size
        expected_checksum = expected_checksum.upper()
        actual_checksum = "SHA-256:" + actual_checksum.upper()
        is_checksum_valid = actual_checksum == expected_checksum

        return is_size_valid, is_checksum_valid

    def __extract_zip(self, archive_filepath, dest_folder):
        try:
            with zipfile.ZipFile(archive_filepath, "r") as zip_ref:
                zip_ref.extractall(dest_folder)
        except:
            self.set_install_error("Error while unpacking {}".format(archive_filepath))
            return

    def __extract_targz(self, archive_filepath, dest_folder):
        try:
            with zipfile.ZipFile(archive_filepath, "r") as zip_ref:
                zip_ref.extractall(dest_folder)
        except:
            self.set_install_error("Error while unpacking {}".format(archive_filepath))
            return

    def __extract_archive(self, archive_filepath, dest_folder):
        """Extracts an archive into a specific folder in arduino-compatable way."""
        name = pathlib.Path(archive_filepath).name
        print("Extracting archive {} into {}".format(name, dest_folder))

        extention = pathlib.Path(archive_filepath).suffix

        if extention == '.zip':
            self.__extract_zip(archive_filepath, dest_folder)
        elif extention == '.gz':
            tar = tarfile.open(archive_filepath, "r:gz")
            tar.extractall(path=dest_folder)
            tar.close()
        else:
            self.set_install_error("Unknown archive extention {}, {}".format(extention, archive_filepath))
            return

        # Move content to the top level
        items = os.listdir(dest_folder)
        if len(items) != 1:
            self.set_install_error("Error in archive {} structure".format(archive_filepath))
            return

        top_folder = None
        for item in items:
            item_path = os.path.join(dest_folder, item)
            if os.path.isdir(item_path):
                top_folder = item_path
                break

        if not top_folder:
            self.set_install_error("Error in archive {} structure".format(archive_filepath))
            return
        
        for item in os.listdir(top_folder):
            src = os.path.join(top_folder, item)
            dst = os.path.join(dest_folder, item)
            print("src: {}".format(src))
            print("dst: {}".format(dst))
            shutil.move(src, dst)

        os.rmdir(top_folder)

    def __install_archive(self, metadata: Archive_Metadata):
        url = metadata.url
        archive_path = os.path.join(self.__temp_folder_path, metadata.archiveFileName)

        self.__download_file(url, archive_path)

        if self.has_error():
            return

        print("{} was successfully downloaded into {}".format(metadata.archiveFileName, archive_path))

        # Verify the checksum and size
        sv, chsv = self.__check_archive(archive_path, metadata.size, metadata.checksum)
        if not sv:
            self.set_install_error(
                "The size {.size} of {} is not valid".format(metadata, metadata.archiveFileName)
            )
            if self.has_error():
                return

        if not chsv:
            self.set_install_error(
                "The checksum {} of {} is not valid".format(metadata.checksum, metadata.archiveFileName)
            )
            if self.has_error():
                return

        # Extract into the destination_directory
        self.__extract_archive(archive_path, metadata.destination_directory)
        if self.has_error():
            return

        print(
            "Successfully extracted archive {} into {}".format(metadata.archiveFileName, metadata.destination_directory)
        )

    def __get_target_info(self):
        # machine = platform.machine()
        system = platform.system()

        if system == "Windows":
            target = "i686-mingw32"
        elif system == "Darwin":
            target = "x86_64-apple-darwin"
        elif system == "Linux":
            target = "x86_64-pc-linux-gnu"
        else:
            target = None

        return target

    def __find_system(self, tool_json):
        host = self.__get_target_info()

        for system_json in tool_json["systems"]:
            if system_json["host"] == host or system_json["host"] == "all":
                return system_json

        return None

    def __find_dependency_tool_system(self, dependency_json, tools_json):
        for tool_json in tools_json:
            if tool_json["name"] == dependency_json["name"]:
                if tool_json["version"] == dependency_json["version"]:
                    return self.__find_system(tool_json)

        return None

    def __install_tool_dependency(self, dependency_json, tools_json):
        tool_system_json = self.__find_dependency_tool_system(
            dependency_json, tools_json
        )
        if tool_system_json == None:
            self.set_install_error(
                "Error! The dependency {} could not be met".format(dependency_json)
            )
            return

        tool_name_path = os.path.join(self.__temp_tools_path, dependency_json["name"])
        tool_version_path = os.path.join(tool_name_path, dependency_json["version"])
        self.__install_archive(
            self.Archive_Metadata(
                tool_system_json["archiveFileName"],
                tool_system_json["url"],
                tool_system_json["checksum"],
                tool_system_json["size"],
                tool_version_path,
            )
        )
        if self.has_error():
            return

    def __install_tools_dependencies(self, dependencies_json, tools_json):
        for dependency_json in dependencies_json:
            self.__install_tool_dependency(dependency_json, tools_json)
            if self.has_error():
                return

    def __install_platform(self, hardware_path, platform_json):
        architecture_path = os.path.join(hardware_path, platform_json["architecture"])
        os.mkdir(architecture_path)
        architecture_version_path = os.path.join(
            architecture_path, platform_json["version"]
        )
        os.mkdir(architecture_version_path)

        self.__install_archive(
            self.Archive_Metadata(
                platform_json["archiveFileName"],
                platform_json["url"],
                platform_json["checksum"],
                platform_json["size"],
                architecture_version_path,
            )
        )
        if self.has_error():
            return

    def __install_index(self):
        url = self.__source_url + self.__index_filename
        index_json_path = os.path.join(self.__temp_folder_path, self.__index_filename)

        self.__download_file(url, index_json_path)
        if self.has_error():
            return

        print("Index file was successfully downloaded into {}".format(index_json_path))

        with open(index_json_path, "r") as file:
            index_json = json.load(file)
            return index_json

    def __install_prepare_temp(self):
        """Create temporary distribution directory"""
        temp_dir = tempfile.gettempdir()

        folder_name = "Rudiron_Distribution"
        self.__temp_folder_path = os.path.join(temp_dir, folder_name)
        print("Temporary folder path: {}".format(self.__temp_folder_path))

        if os.path.exists(self.__temp_folder_path):
            shutil.rmtree(self.__temp_folder_path)

        os.makedirs(self.__temp_folder_path)

    def __install_remove_temp(self):
        """Deletes temporary distribution directory"""
        if os.path.exists(self.__temp_folder_path):
            shutil.rmtree(self.__temp_folder_path)

    def install(self):
        """
        Complete installation of Rudiron distribution
        """
        self.__install_prepare_temp()

        index_json = self.__install_index()
        if index_json == None:
            return

        package = index_json["packages"][0]

        maintainer_json = package["maintainer"]
        self.__temp_maintainer_path = os.path.join(
            self.__temp_folder_path, maintainer_json
        )
        os.mkdir(self.__temp_maintainer_path)

        self.__temp_tools_path = os.path.join(self.__temp_maintainer_path, "tools")
        os.mkdir(self.__temp_tools_path)

        self.__temp_hardware_path = os.path.join(
            self.__temp_maintainer_path, "hardware"
        )
        os.mkdir(self.__temp_hardware_path)

        tools_json = package["tools"]
        for platform_json in package["platforms"]:
            self.__install_platform(self.__temp_hardware_path, platform_json)
            if self.has_error():
                return

            self.__install_tools_dependencies(
                platform_json["toolsDependencies"], tools_json
            )
            if self.has_error():
                return

        if os.path.exists(self.__get_arduino_maintainer_path()):
            shutil.rmtree(self.__get_arduino_maintainer_path())

        shutil.move(self.__temp_maintainer_path, self.__get_arduino_maintainer_path())

        if self.has_error():
            return
        else:
            self.__install_remove_temp()
            return

    def uninstall(self):
        """
        Complete uninstallation of Rudiron distribution
        """
        if self.is_installed():
            print("Existing Rudiron distribution is found and will be deleted")
            shutil.rmtree(self.__get_arduino_maintainer_path())

    def install_vscode_extentions(self):
        for extention in self.__vscode_extentions:
            result_code = subprocess.call(
                "code --install-extension \"{}\" --force".format(extention), shell=True
            )

            if result_code != 0:
                self.set_install_error(
                    "Error! The extention {} could not be installed".format(extention)
                )
                return

    def update(self):
        """
        Partial reinstallation, includes only platform upgrades.
        """
        self.__install_prepare_temp()

        index_json = self.__install_index()
        if index_json == None:
            return

        package = index_json["packages"][0]

        maintainer_json = package["maintainer"]
        self.__temp_maintainer_path = os.path.join(
            self.__temp_folder_path, maintainer_json
        )
        os.mkdir(self.__temp_maintainer_path)

        self.__temp_hardware_path = os.path.join(
            self.__temp_maintainer_path, "hardware"
        )
        os.mkdir(self.__temp_hardware_path)

        for platform_json in package["platforms"]:
            self.__install_platform(self.__temp_hardware_path, platform_json)
            if self.has_error():
                return

        maintainer_hardware_path = os.path.join(self.__get_arduino_maintainer_path(), "hardware")
        if os.path.exists(maintainer_hardware_path):
            shutil.rmtree(maintainer_hardware_path)

        shutil.move(self.__temp_hardware_path, maintainer_hardware_path)

        if self.has_error():
            return
        else:
            self.__install_remove_temp()
            return
